/**
 * This is a collection object that contains items of an certain entity.
 * This collection "lives" in the sense that its items are automatically
 * updated, added and removed. When such a change happens, an event is triggered* you can listen on.
 */
import { ReplaySubject, Subject, TeardownLogic } from 'rxjs';
import { IdInterface } from './contract';
import { ClassType } from '@deepkit/core';
import { EntitySubject } from './core';
export declare type FilterParameters = {
    [name: string]: any | undefined;
};
export interface CollectionBatchStart {
    type: 'batch/start';
}
export interface CollectionBatchEnd {
    type: 'batch/end';
}
export interface CollectionAdd {
    type: 'add';
    item: any;
}
export interface CollectionRemove {
    type: 'remove';
    id: string | number;
}
export interface CollectionSetSort {
    type: 'sort';
    ids: (string | number)[];
}
export interface CollectionRemoveMany {
    type: 'removeMany';
    ids: (string | number)[];
}
export interface CollectionSet {
    type: 'set';
    items: any[];
}
export declare type CollectionEvent = CollectionBatchStart | CollectionBatchEnd | CollectionAdd | CollectionSetSort | CollectionRemove | CollectionRemoveMany | CollectionSet;
export declare type CollectionSortDirection = 'asc' | 'desc';
export interface CollectionSort {
    field: string;
    direction: CollectionSortDirection;
}
export interface CollectionPaginationEventApplyFinished {
    type: 'server:apply/finished';
}
export interface CollectionPaginationEventInternalChange {
    type: 'internal_server_change';
}
export interface CollectionPaginationEventChange {
    type: 'server:change';
    order: {
        field: string;
        direction: 'asc' | 'desc';
    }[];
    parameters: FilterParameters;
    itemsPerPage: number;
    total: number;
    page: number;
}
export interface CollectionPaginationEventApply {
    type: 'apply';
}
export interface CollectionPaginationEventClientApply {
    type: 'client:apply';
}
export declare type CollectionPaginationEvent = CollectionPaginationEventApplyFinished | CollectionPaginationEventChange | CollectionPaginationEventInternalChange | CollectionPaginationEventClientApply | CollectionPaginationEventApply;
export declare class CollectionPagination<T extends IdInterface> {
    private collection;
    readonly event: Subject<CollectionPaginationEvent>;
    _parameters: FilterParameters;
    constructor(collection: Collection<T>);
    protected page: number;
    protected total: number;
    protected itemsPerPage: number;
    protected sort: CollectionSort[];
    protected active: boolean;
    protected applyPromise?: Promise<void>;
    protected applyPromiseResolver?: () => void;
    setParameters(values?: FilterParameters): CollectionPagination<T>;
    resetParameters(): CollectionPagination<T>;
    setParameter(name: string, value?: any): CollectionPagination<T>;
    getParameter(name: string): any;
    getParameters(): FilterParameters;
    getHash(): string;
    getPagingHash(): string;
    getParametersHash(): string;
    setPage(page: number): CollectionPagination<T>;
    hasSort(): boolean;
    getSort(): CollectionSort[];
    setSort(order: CollectionSort[]): CollectionPagination<T>;
    orderByField(field: string, direction?: CollectionSortDirection): CollectionPagination<T>;
    /**
     * Sends current pagination setting to the server and refreshes the collection if necessary.
     */
    apply(): Promise<void>;
    /**
     * Triggered from the server when the apply() finished. Doesn't matter if we got actual updates or not. It's important
     * to distinguish because collection.nextStateChange is not reliable enough
     * (as entity updates could happen between apply and applyFinish, which would trigger nextStateChange).
     *
     * @private
     */
    _applyFinished(): void;
    setTotal(total: number): CollectionPagination<T>;
    setItemsPerPage(items: number): CollectionPagination<T>;
    isActive(): boolean;
    /**
     * It's not possible to activate this later on. This is set on the server side only.
     */
    _activate(): CollectionPagination<T>;
    getPage(): number;
    getItemsPerPage(): number;
    getTotal(): number;
    getPages(): number;
    isPageValid(): boolean;
}
export interface CollectionEntitySubjectFetcher {
    fetch<T extends IdInterface>(classType: ClassType<T>, id: string | number): EntitySubject<T>;
}
export declare class Collection<T extends IdInterface> extends ReplaySubject<T[]> {
    readonly classType: ClassType<T>;
    readonly event: Subject<CollectionEvent>;
    readonly removed: Subject<T>;
    readonly added: Subject<T>;
    protected readonly teardowns: TeardownLogic[];
    protected items: T[];
    protected itemsMapped: {
        [id: string]: T;
    };
    readonly deepChange: Subject<T>;
    protected nextChange?: Subject<void>;
    readonly pagination: CollectionPagination<T>;
    protected batchActive: boolean;
    protected batchNeedLoaded: boolean;
    protected entitySubjectFetcher?: CollectionEntitySubjectFetcher;
    readonly entitySubjects: {
        [id: string]: EntitySubject<T>;
    };
    constructor(classType: ClassType<T>);
    getEntitySubject(idOrItem: string | T): EntitySubject<T>;
    has(id: string | number): boolean;
    get(id: string | number): T | undefined;
    batchStart(): void;
    batchEnd(): void;
    /**
     * Resolves when next change happened.
     */
    get nextStateChange(): Promise<void>;
    /**
     * Unsubscribe from the backend stream.
     */
    unsubscribe(): Promise<void>;
    addTeardown(teardown: TeardownLogic): void;
    index(item: T): number;
    /**
     * Returns the page zero-based of the current item.
     */
    getPageOf(item: T, itemsPerPage?: number): number;
    reset(): void;
    all(): T[];
    count(): number;
    ids(): (string | number)[];
    empty(): boolean;
    map(): {
        [id: string]: T;
    };
    loaded(): void;
    seItem(id: string | number, item: T): void;
    set(items: T[], withEvent?: boolean): void;
    setSort(ids: (string | number)[]): void;
    removeMany(ids: (string | number)[], withEvent?: boolean): void;
    add(item: T, withEvent?: boolean): void;
    remove(id: string | number, withEvent?: boolean): void;
}
declare type JSONObject<T> = Partial<T> & IdInterface;
export declare class JSONObjectCollection<T extends JSONObject<T>> extends Collection<T> {
}
export {};
